/*
								+----------------------------------+
								|                                  |
								|  ***   Catmull-Rom spline   ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|           CatmullRom.h           |
								|                                  |
								+----------------------------------+
*/

#ifndef __CATMULL_ROM_SPLINE_INCLUDED
#define __CATMULL_ROM_SPLINE_INCLUDED

#include "Spline.h"

/*
 *	class CCatmullRomSpline
 *		- simple Ctamull - Rom spline class
 */
class CCatmullRomSpline : public CSpline<Vector3f> {
public:
	/*
	 *	CCatmullRomSpline::CCatmullRomSpline(const std::vector<Vector3f> &r_point_list)
	 *		- default constructor
	 *		- r_point_list is list of curve points, it's not copied, only referenced
	 *		  and must remain unchanged (arc lengths precalculated in constructor)
	 *		- note catmull-rom splines have segments from second to last but one point
	 *		  so it may be useful to duplicate first and last points
	 */
	CCatmullRomSpline(const std::vector<Vector3f> &r_point_list);

	/*
	 *	virtual int CCatmullRomSpline::n_Arc_Num() const
	 *		- return number of spline arcs
	 */
	virtual int n_Arc_Num() const;

	/*
	 *	virtual Vector3f CCatmullRomSpline::v_Arc_Position(float f_t, int n_arc) const
	 *		- returns point at position f_t along curve arc n_arc
	 */
	virtual Vector3f v_Arc_Position(float f_t, int n_arc) const;

	/*
	 *	virtual Vector3f CCatmullRomSpline::v_Arc_Derivative(float f_t, int n_arc) const
	 *		- returns tangent at position f_t along curve arc n_arc
	 */
	virtual Vector3f v_Arc_Derivative(float f_t, int n_arc) const;
};

#endif //__CATMULL_ROM_SPLINE_INCLUDED
